import functools
from typing import List, Any
from urllib.parse import urljoin

import zeep.helpers
from zeep import Client
from zeep.xsd import ComplexType

from .helpers import modify_url_for_urljoin
from .model import Identity
from .schema import load_identity, dump_identity


def create_epix_client(base_url: str) -> Client:
    """
    Creates a Zeep client for the E-PIX SOAP interface hosted at the given URL.

    :param base_url: URL where the E-PIX service is hosted
    :return: Zeep client
    """
    wsdl_url = urljoin(modify_url_for_urljoin(base_url), "epixService?wsdl")
    return Client(wsdl_url)


def get_mpi_request_type(client: Client) -> ComplexType:
    """
    Returns the MPI request type as defined by the SOAP interface.

    :param client: Zeep client with E-PIX SOAP methods
    :return: SOAP MPI request type
    """
    return client.get_type("ns0:mpiRequestDTO")


def _get_mpi_from_mpi_batch_response_entry(entry: Any) -> str:
    return entry.value.person.mpiId.value


def request_mpi_batch(client: Client, domain: str, source: str, identities: List[Identity]) -> List[str]:
    """
    Submits many identities to the MPI for creation. Returns the identifiers assigned to these identities.

    :param client: Zeep client with E-PIX SOAP methods
    :param domain: Data domain
    :param source: Data source
    :param identities: Identities to insert
    :return: List of identifiers
    """
    create_soap_mpi_batch_request = get_mpi_request_type(client)

    entries = [
        dump_identity(i) for i in identities
    ]

    request = create_soap_mpi_batch_request(
        domainName=domain,
        sourceName=source,
        requestEntries=entries
    )

    response = client.service.requestMPIBatch(request)

    if response is None:
        return []

    return [
        _get_mpi_from_mpi_batch_response_entry(e) for e in response.entry
    ]


def _get_identity_id_from_person_for_domain_response_entry(entry: Any) -> int:
    """
    Returns the MPI ID for an identity in a domain response entry.

    :param entry: Entry to extract MPI ID from
    :return: MPI ID
    """
    return entry.referenceIdentity.identityId


def get_identity_ids_in_domain(client: Client, domain: str) -> List[int]:
    """
    Fetches all identity IDs present in the given domain.

    :param client: Zeep client with E-PIX SOAP methods
    :param domain: Data domain
    :return: List of identity IDs
    """
    response = client.service.getPersonsForDomain(domain)
    return [
        _get_identity_id_from_person_for_domain_response_entry(e) for e in response
    ]


def delete_identity(client: Client, identity_id: int) -> None:
    """
    Deletes the identity with the given ID.

    :param client: Zeep client with E-PIX SOAP methods
    :param identity_id: ID of the identity to delete
    """
    client.service.deleteIdentity(identity_id)


def deactivate_identity(client: Client, identity_id: int) -> None:
    """
    Deactivates the identity with the given ID.

    :param client: Zeep client with E-PIX SOAP methods
    :param identity_id: ID of the identity to deactivate
    """
    client.service.deactivateIdentity(identity_id)


def resolve_person_for_mpi(client: Client, domain: str, mpi: str) -> Identity:
    """
    Looks up an identifier in the MPI and returns the corresponding personal entry, if present.

    :param client: Zeep client with E-PIX SOAP methods
    :param domain: Data domain
    :param mpi: MPI record identifier
    :return: Personal data associated with the given identifier
    """
    person_data = client.service.getPersonByMPI(domain, mpi)
    person_data_id = person_data["referenceIdentity"]

    return load_identity(zeep.helpers.serialize_object(person_data_id, dict))


class EPIXClient:

    def __init__(self, base_url: str):
        """
        Creates a convenience wrapper around selected E-PIX SOAP functions.

        :param base_url: URL where the E-PIX service is hosted
        """
        client = create_epix_client(base_url)
        self.resolve_person_for_mpi = functools.partial(resolve_person_for_mpi, client)
        self.request_mpi_batch = functools.partial(request_mpi_batch, client)
        self.get_mpis_in_domain = functools.partial(get_identity_ids_in_domain, client)
        self.delete_identity = functools.partial(delete_identity, client)
        self.deactivate_identity = functools.partial(deactivate_identity, client)



import functools
from typing import List, Dict
from urllib.parse import urljoin

from zeep import Client

from .helpers import modify_url_for_urljoin


def create_gpas_client(base_url: str) -> Client:
    """
    Creates a Zeep client for the gPAS SOAP interface hosted at the given URL.

    :param base_url: URL where the gPAS service is hosted
    :return: Zeep client
    """
    wsdl_url = urljoin(modify_url_for_urljoin(base_url), "gpasService?wsdl")
    return Client(wsdl_url)


def resolve_pseudonym(client: Client, domain: str, pseudonym: str) -> str:
    """
    Looks up a domain-specific pseudonym and returns the corresponding identifier that is associated with it.

    :param client: Zeep client with gPAS SOAP methods
    :param domain: Data domain
    :param pseudonym: Psudonym to resolve
    :return: Identifier that is associated with the given pseudonym
    """
    return client.service.getValueFor(pseudonym, domain)


def get_or_create_pseudonyms(client: Client, domain: str, values: List[str]) -> Dict[str, str]:
    """
    Gets or creates pseudonyms for the given values in the provided domain.

    :param client: Zeep client with gPAS SOAP methods
    :param domain: Data domain
    :param values: Values to create pseudonyms for
    :return: Mapping of values to their pseudonyms
    """
    resp = client.service.getOrCreatePseudonymForList(values, domain)

    return {
        entry.key: entry.value for entry in resp
    }


def delete_entries(client: Client, domain: str, values: List[str]) -> bool:
    """
    Deletes entries from the database. These are not pseudonyms but rather the values that were used to
    create the pseudonyms.

    :param client: Zeep client with gPAS SOAP methods
    :param domain: Data domain
    :param values: Values to delete
    :return: ``True`` if all values have been deleted successfully, ``False`` otherwise
    """
    if len(values) == 0:
        return True

    resp = client.service.deleteEntries(values, domain)

    for entry in resp:
        if entry.value != "SUCCESS":
            return False

    return True


class GPASClient:

    def __init__(self, base_url: str):
        """
        Creates a convenience wrapper around selected gPAS SOAP functions.

        :param base_url: URL where the gPAS service is hosted
        """
        client = create_gpas_client(base_url)
        self.resolve_pseudonym = functools.partial(resolve_pseudonym, client)
        self.get_or_create_pseudonyms = functools.partial(get_or_create_pseudonyms, client)
        self.delete_entries = functools.partial(delete_entries, client)



import urllib.parse


def modify_url_for_urljoin(url: str):
    """
    Removes params, query and fragment from the URL and adds a trailing slash to the path if it doesn't exist yet.

    :param url: URL to modify
    :return: Modified URL
    """
    scheme, host, path, _, _, _ = urllib.parse.urlparse(url)

    if not host.endswith("/"):
        host = host + "/"

    return urllib.parse.urlunparse((scheme, host, path, "", "", ""))



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List


@dataclass
class Contact:
    city: Optional[str] = None
    """City of residence (e.g. Berlin)"""
    country: Optional[str] = None
    """Country of residence (e.g. Germany)"""
    country_code: Optional[str] = None
    """Country code (e.g. DE)"""
    district: Optional[str] = None
    """District (e.g. Spandau)"""
    email: Optional[str] = None
    """Email address (e.g. anna.schmidt@beispiel.de)"""
    external_date: Optional[datetime] = field(default=None)
    """Custom external date (e.g. 2019-06-27)"""
    municipality_key: Optional[str] = None
    """Municipality key (e.g. 11000000)"""
    phone: Optional[str] = None
    """Phone number (e.g. 030/123456789)"""
    state: Optional[str] = None
    """State (e.g. Berlin)"""
    street: Optional[str] = None
    """Street (e.g. Spandauer Damm)"""
    zip_code: Optional[str] = None
    """Zip Code (e.g. 13593)"""

@dataclass
class Identity:
    birth_date: datetime = field(default=None)
    """Birth date (e.g. 2019-04-30)"""
    birth_place: Optional[str] = None
    """Birth place (e.g. Berlin)"""
    civil_status: Optional[str] = None
    """Civil status (e.g. Single)"""
    degree: Optional[str] = None
    """Degree (e.g. Higher Education)"""
    external_date: Optional[datetime] = field(default=None)
    """Custom external date (e.g. 2019-06-27)"""
    first_name: Optional[str] = None
    """First name (e.g. Anna)"""
    gender: Optional[str] = None
    """Gender (M = male, F = female, O = other, U = unknown, X = divers)"""
    last_name: Optional[str] = None
    """Last name (e.g. Lea)"""
    middle_name: Optional[str] = None
    """Middle name(s) (e.g. Schmidt)"""
    mother_tongue: Optional[str] = None
    """Mother tongue (e.g. German)"""
    mothers_maiden_name: Optional[str] = None
    """Mothers' maiden name (e.g. Müller)"""
    nationality: Optional[str] = None
    """Nationality (e.g. German)"""
    prefix: Optional[str] = None
    """Prefix (e.g. of)"""
    race: Optional[str] = None
    """Race (e.g. Caucasian)"""
    religion: Optional[str] = None
    """Religion (e.g. Christianity)"""
    suffix: Optional[str] = None
    """Suffix (e.g. B.Sc.)"""
    # custom fields
    value1: Optional[str] = None
    """Custom field (up to 50 chars)"""
    value2: Optional[str] = None
    """Custom field (up to 50 chars)"""
    value3: Optional[str] = None
    """Custom field (up to 50 chars)"""
    value4: Optional[str] = None
    """Custom field (up to 50 chars)"""
    value5: Optional[str] = None
    """Custom field (up to 50 chars)"""
    value6: Optional[str] = None
    """Custom field (up to 255 chars)"""
    value7: Optional[str] = None
    """Custom field (up to 255 chars)"""
    value8: Optional[str] = None
    """Custom field (up to 1.000 chars)"""
    value9: Optional[str] = None
    """Custom field (up to 1.000 chars)"""
    value10: Optional[str] = None
    """Custom field (up to 10.000 chars)"""
    contacts: List[Contact] = field(default_factory=list)
    """List of contact information"""



import datetime
from typing import Any, Optional, Mapping, Dict

from marshmallow import Schema, fields, post_load, EXCLUDE, ValidationError

from .model import Identity, Contact


class DateTimeObjField(fields.Field):

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        if value is None:
            return None

        return value

    def _deserialize(self, value: Any, attr: Optional[str], data: Optional[Mapping[str, Any]], **kwargs):
        if not isinstance(value, datetime.datetime):
            raise ValidationError("Date time object must be an instance of datetime.datetime")

        return value


class ContactSchema(Schema):
    city = fields.Str(missing=None)
    country = fields.Str(missing=None)
    country_code = fields.Str(data_key="countryCode", missing=None)
    district = fields.Str(missing=None)
    email = fields.Str(missing=None)
    external_date = DateTimeObjField(data_key="externalDate", missing=None)
    municipality_key = fields.Str(data_key="municipalityKey", missing=None)
    phone = fields.Str(missing=None)
    state = fields.Str(missing=None)
    street = fields.Str(missing=None)
    zip_code = fields.Str(data_key="zipCode", missing=None)

    @post_load
    def make_contact(self, data, **kwargs):
        return Contact(**data)

    class Meta:
        unknown = EXCLUDE


class IdentitySchema(Schema):
    first_name = fields.Str(data_key="firstName", missing=None)
    last_name = fields.Str(data_key="lastName", missing=None)
    gender = fields.Str(missing=None)
    birth_date = DateTimeObjField(data_key="birthDate", missing=None)
    birth_place = fields.Str(data_key="birthPlace", missing=None)
    civil_status = fields.Str(data_key="civilStatus", missing=None)
    degree = fields.Str(missing=None)
    external_date = DateTimeObjField(data_key="externalDate", missing=None)
    middle_name = fields.Str(data_key="middleName", missing=None)
    mother_tongue = fields.Str(data_key="motherTongue", missing=None)
    mothers_maiden_name = fields.Str(data_key="mothersMaidenName", missing=None)
    nationality = fields.Str(missing=None)
    prefix = fields.Str(missing=None)
    race = fields.Str(missing=None)
    religion = fields.Str(missing=None)
    suffix = fields.Str(missing=None)
    value1 = fields.Str(missing=None)
    value2 = fields.Str(missing=None)
    value3 = fields.Str(missing=None)
    value4 = fields.Str(missing=None)
    value5 = fields.Str(missing=None)
    value6 = fields.Str(missing=None)
    value7 = fields.Str(missing=None)
    value8 = fields.Str(missing=None)
    value9 = fields.Str(missing=None)
    value10 = fields.Str(missing=None)
    contacts = fields.List(fields.Nested(ContactSchema), missing=[])

    @post_load
    def make_identity(self, data, **kwargs):
        return Identity(**data)

    class Meta:
        unknown = EXCLUDE


def load_identity(data: Dict[str, Any]) -> Identity:
    return IdentitySchema().load(data)


def dump_identity(identity: Identity) -> Dict[str, Any]:
    return IdentitySchema().dump(identity)


def load_contact(data: Dict[str, Any]) -> Identity:
    return ContactSchema().load(data)


def dump_contact(contact: Contact) -> Dict[str, Any]:
    return ContactSchema().dump(contact)



from setuptools import setup


def readme():
    with open("README.rst", mode="r", encoding="utf-8") as f:
        return f.read()


setup(name="mdsmosaic",
      version="0.1.2",
      description="Client utilities for the MOSAIC suite",
      long_description=readme(),
      url="https://hub.docker.com/mds4ul",
      author="Maximilian Jugl",
      author_email="jugl@hs-mittweida.de",
      classifiers=[
          "Development Status :: 3 - Alpha",
          "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
          "Natural Language :: English",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Software Development :: Libraries",
          "Typing :: Typed"
      ],
      license="GPL-3.0-or-later",
      packages=["mdsmosaic"],
      install_requires=[
          "marshmallow",
          "zeep"
      ],
      zip_safe=False)
